4" * This file is part of FFmpeg.\n"
6" * FFmpeg is free software; you can redistribute it and/or\n"
7" * modify it under the terms of the GNU Lesser General Public\n"
8" * License as published by the Free Software Foundation; either\n"
9" * version 2.1 of the License, or (at your option) any later version.\n"
11" * FFmpeg is distributed in the hope that it will be useful,\n"
12" * but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
13" * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU\n"
14" * Lesser General Public License for more details.\n"
16" * You should have received a copy of the GNU Lesser General Public\n"
17" * License along with FFmpeg; if not, write to the Free Software\n"
18" * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA\n"
20" * Copyright (C) 2000, Intel Corporation, all rights reserved.\n"
21" * Copyright (C) 2013, OpenCV Foundation, all rights reserved.\n"
22" * Third party copyrights are property of their respective owners.\n"
24" * Redistribution and use in source and binary forms, with or without modification,\n"
25" * are permitted provided that the following conditions are met:\n"
27" * * Redistribution's of source code must retain the above copyright notice,\n"
28" * this list of conditions and the following disclaimer.\n"
30" * * Redistribution's in binary form must reproduce the above copyright notice,\n"
31" * this list of conditions and the following disclaimer in the documentation\n"
32" * and/or other materials provided with the distribution.\n"
34" * * The name of the copyright holders may not be used to endorse or promote products\n"
35" * derived from this software without specific prior written permission.\n"
37" * This software is provided by the copyright holders and contributors \"as is\" and\n"
38" * any express or implied warranties, including, but not limited to, the implied\n"
39" * warranties of merchantability and fitness for a particular purpose are disclaimed.\n"
40" * In no event shall the Intel Corporation or contributors be liable for any direct,\n"
41" * indirect, incidental, special, exemplary, or consequential damages\n"
42" * (including, but not limited to, procurement of substitute goods or services;\n"
43" * loss of use, data, or profits; or business interruption) however caused\n"
44" * and on any theory of liability, whether in contract, strict liability,\n"
45" * or tort (including negligence or otherwise) arising in any way out of\n"
46" * the use of this software, even if advised of the possibility of such damage.\n"
49"#define HARRIS_THRESHOLD 3.0f\n"
50"// Block size over which to compute harris response\n"
52"// Note that changing this will require fiddling with the local array sizes in\n"
54"#define HARRIS_RADIUS 2\n"
55"#define DISTANCE_THRESHOLD 80\n"
57"// Sub-pixel refinement window for feature points\n"
58"#define REFINE_WIN_HALF_W 5\n"
59"#define REFINE_WIN_HALF_H 5\n"
60"#define REFINE_WIN_W 11 // REFINE_WIN_HALF_W * 2 + 1\n"
61"#define REFINE_WIN_H 11\n"
63"// Non-maximum suppression window size\n"
64"#define NONMAX_WIN 30\n"
65"#define NONMAX_WIN_HALF 15 // NONMAX_WIN / 2\n"
67"typedef struct PointPair {\n"
74"typedef struct SmoothedPointPair {\n"
75" // Non-smoothed point in current frame\n"
77" // Smoothed point in current frame\n"
79"} SmoothedPointPair;\n"
81"typedef struct MotionVector {\n"
83" // Used to mark vectors as potential outliers\n"
84" int should_consider;\n"
87"const sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE |\n"
88" CLK_ADDRESS_CLAMP_TO_EDGE |\n"
89" CLK_FILTER_NEAREST;\n"
91"const sampler_t sampler_linear = CLK_NORMALIZED_COORDS_FALSE |\n"
92" CLK_ADDRESS_CLAMP_TO_EDGE |\n"
93" CLK_FILTER_LINEAR;\n"
95"const sampler_t sampler_linear_mirror = CLK_NORMALIZED_COORDS_TRUE |\n"
96" CLK_ADDRESS_MIRRORED_REPEAT |\n"
97" CLK_FILTER_LINEAR;\n"
99"// Writes to a 1D array at loc, treating it as a 2D array with the same\n"
100"// dimensions as the global work size.\n"
101"static void write_to_1d_arrf(__global float *buf, int2 loc, float val) {\n"
102" buf[loc.x + loc.y * get_global_size(0)] = val;\n"
105"static void write_to_1d_arrul8(__global ulong8 *buf, int2 loc, ulong8 val) {\n"
106" buf[loc.x + loc.y * get_global_size(0)] = val;\n"
109"static void write_to_1d_arrvec(__global MotionVector *buf, int2 loc, MotionVector val) {\n"
110" buf[loc.x + loc.y * get_global_size(0)] = val;\n"
113"static void write_to_1d_arrf2(__global float2 *buf, int2 loc, float2 val) {\n"
114" buf[loc.x + loc.y * get_global_size(0)] = val;\n"
117"static ulong8 read_from_1d_arrul8(__global const ulong8 *buf, int2 loc) {\n"
118" return buf[loc.x + loc.y * get_global_size(0)];\n"
121"static float2 read_from_1d_arrf2(__global const float2 *buf, int2 loc) {\n"
122" return buf[loc.x + loc.y * get_global_size(0)];\n"
125"// Returns the grayscale value at the given point.\n"
126"static float pixel_grayscale(__read_only image2d_t src, int2 loc) {\n"
127" float4 pixel = read_imagef(src, sampler, loc);\n"
128" return (pixel.x + pixel.y + pixel.z) / 3.0f;\n"
131"static float convolve(\n"
132" __local const float *grayscale,\n"
139" // These loops touch each pixel surrounding loc as well as loc itself\n"
140" for (int i = 1, i2 = 0; i >= -1; --i, ++i2) {\n"
141" for (int j = -1, j2 = 0; j <= 1; ++j, ++j2) {\n"
142" ret += mask[i2][j2] * grayscale[(local_idx_x + 3 + j) + (local_idx_y + 3 + i) * 14];\n"
149"// Sums dx * dy for all pixels within radius of loc\n"
150"static float sum_deriv_prod(\n"
151" __local const float *grayscale,\n"
152" float mask_x[3][3],\n"
153" float mask_y[3][3]\n"
157" for (int i = HARRIS_RADIUS; i >= -HARRIS_RADIUS; --i) {\n"
158" for (int j = -HARRIS_RADIUS; j <= HARRIS_RADIUS; ++j) {\n"
159" ret += convolve(grayscale, get_local_id(0) + j, get_local_id(1) + i, mask_x) *\n"
160" convolve(grayscale, get_local_id(0) + j, get_local_id(1) + i, mask_y);\n"
167"// Sums d<>^2 (determined by mask) for all pixels within radius of loc\n"
168"static float sum_deriv_pow(__local const float *grayscale, float mask[3][3])\n"
172" for (int i = HARRIS_RADIUS; i >= -HARRIS_RADIUS; --i) {\n"
173" for (int j = -HARRIS_RADIUS; j <= HARRIS_RADIUS; ++j) {\n"
174" float deriv = convolve(grayscale, get_local_id(0) + j, get_local_id(1) + i, mask);\n"
175" ret += deriv * deriv;\n"
182"// Fills a box with the given radius and pixel around loc\n"
183"static void draw_box(__write_only image2d_t dst, int2 loc, float4 pixel, int radius)\n"
185" for (int i = -radius; i <= radius; ++i) {\n"
186" for (int j = -radius; j <= radius; ++j) {\n"
190" // Clamp to avoid writing outside image bounds\n"
191" clamp(loc.x + i, 0, get_image_dim(dst).x - 1),\n"
192" clamp(loc.y + j, 0, get_image_dim(dst).y - 1)\n"
200"// Converts the src image to grayscale\n"
201"__kernel void grayscale(\n"
202" __read_only image2d_t src,\n"
203" __write_only image2d_t grayscale\n"
205" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
206" write_imagef(grayscale, loc, (float4)(pixel_grayscale(src, loc), 0.0f, 0.0f, 1.0f));\n"
209"// This kernel computes the harris response for the given grayscale src image\n"
210"// within the given radius and writes it to harris_buf\n"
211"__kernel void harris_response(\n"
212" __read_only image2d_t grayscale,\n"
213" __global float *harris_buf\n"
215" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
217" if (loc.x > get_image_width(grayscale) - 1 || loc.y > get_image_height(grayscale) - 1) {\n"
218" write_to_1d_arrf(harris_buf, loc, 0);\n"
222" float scale = 1.0f / ((1 << 2) * HARRIS_RADIUS * 255.0f);\n"
224" float sobel_mask_x[3][3] = {\n"
230" float sobel_mask_y[3][3] = {\n"
236" // 8 x 8 local work + 3 pixels around each side (needed to accommodate for the\n"
237" // block size radius of 2)\n"
238" __local float grayscale_data[196];\n"
240" int idx = get_group_id(0) * get_local_size(0);\n"
241" int idy = get_group_id(1) * get_local_size(1);\n"
243" for (int i = idy - 3, it = 0; i < idy + (int)get_local_size(1) + 3; i++, it++) {\n"
244" for (int j = idx - 3, jt = 0; j < idx + (int)get_local_size(0) + 3; j++, jt++) {\n"
245" grayscale_data[jt + it * 14] = read_imagef(grayscale, sampler, (int2)(j, i)).x;\n"
249" barrier(CLK_LOCAL_MEM_FENCE);\n"
251" float sumdxdy = sum_deriv_prod(grayscale_data, sobel_mask_x, sobel_mask_y);\n"
252" float sumdx2 = sum_deriv_pow(grayscale_data, sobel_mask_x);\n"
253" float sumdy2 = sum_deriv_pow(grayscale_data, sobel_mask_y);\n"
255" float trace = sumdx2 + sumdy2;\n"
256" // r = det(M) - k(trace(M))^2\n"
257" // k usually between 0.04 to 0.06\n"
258" float r = (sumdx2 * sumdy2 - sumdxdy * sumdxdy) - 0.04f * (trace * trace) * pown(scale, 4);\n"
260" // Threshold the r value\n"
261" harris_buf[loc.x + loc.y * get_image_width(grayscale)] = r * step(HARRIS_THRESHOLD, r);\n"
264"// Gets a patch centered around a float coordinate from a grayscale image using\n"
265"// bilinear interpolation\n"
266"static void get_rect_sub_pix(\n"
267" __read_only image2d_t grayscale,\n"
273" float2 offset = ((float2)(size_x, size_y) - 1.0f) * 0.5f;\n"
275" for (int i = 0; i < size_y; i++) {\n"
276" for (int j = 0; j < size_x; j++) {\n"
277" buffer[i * size_x + j] = read_imagef(\n"
280" (float2)(j, i) + center - offset\n"
286"// Refines detected features at a sub-pixel level\n"
288"// This function is ported from OpenCV\n"
289"static float2 corner_sub_pix(\n"
290" __read_only image2d_t grayscale,\n"
294" float2 init = feature;\n"
295" int src_width = get_global_size(0);\n"
296" int src_height = get_global_size(1);\n"
298" const int max_iters = 40;\n"
299" const float eps = 0.001f * 0.001f;\n"
304" float subpix[(REFINE_WIN_W + 2) * (REFINE_WIN_H + 2)];\n"
305" const float flt_epsilon = 0x1.0p-23f;\n"
308" float2 feature_tmp;\n"
309" float a = 0, b = 0, c = 0, bb1 = 0, bb2 = 0;\n"
311" get_rect_sub_pix(grayscale, subpix, REFINE_WIN_W + 2, REFINE_WIN_H + 2, feature);\n"
312" float *subpix_ptr = subpix;\n"
313" subpix_ptr += REFINE_WIN_W + 2 + 1;\n"
315" // process gradient\n"
316" for (i = 0, k = 0; i < REFINE_WIN_H; i++, subpix_ptr += REFINE_WIN_W + 2) {\n"
317" float py = i - REFINE_WIN_HALF_H;\n"
319" for (j = 0; j < REFINE_WIN_W; j++, k++) {\n"
320" float m = mask[k];\n"
321" float tgx = subpix_ptr[j + 1] - subpix_ptr[j - 1];\n"
322" float tgy = subpix_ptr[j + REFINE_WIN_W + 2] - subpix_ptr[j - REFINE_WIN_W - 2];\n"
323" float gxx = tgx * tgx * m;\n"
324" float gxy = tgx * tgy * m;\n"
325" float gyy = tgy * tgy * m;\n"
326" float px = j - REFINE_WIN_HALF_W;\n"
332" bb1 += gxx * px + gxy * py;\n"
333" bb2 += gxy * px + gyy * py;\n"
337" float det = a * c - b * b;\n"
338" if (fabs(det) <= flt_epsilon * flt_epsilon) {\n"
342" // 2x2 matrix inversion\n"
343" float scale = 1.0f / det;\n"
344" feature_tmp.x = (float)(feature.x + (c * scale * bb1) - (b * scale * bb2));\n"
345" feature_tmp.y = (float)(feature.y - (b * scale * bb1) + (a * scale * bb2));\n"
346" err = dot(feature_tmp - feature, feature_tmp - feature);\n"
348" feature = feature_tmp;\n"
349" if (feature.x < 0 || feature.x >= src_width || feature.y < 0 || feature.y >= src_height) {\n"
352" } while (++iter < max_iters && err > eps);\n"
354" // Make sure new point isn't too far from the initial point (indicates poor convergence)\n"
355" if (fabs(feature.x - init.x) > REFINE_WIN_HALF_W || fabs(feature.y - init.y) > REFINE_WIN_HALF_H) {\n"
362"// Performs non-maximum suppression on the harris response and writes the resulting\n"
363"// feature locations to refined_features.\n"
365"// Assumes that refined_features and the global work sizes are set up such that the image\n"
366"// is split up into a grid of 32x32 blocks where each block has a single slot in the\n"
367"// refined_features buffer. This kernel finds the best corner in each block (if the\n"
368"// block has any) and writes it to the corresponding slot in the buffer.\n"
370"// If subpixel_refine is true, the features are additionally refined at a sub-pixel\n"
371"// level for increased precision.\n"
372"__kernel void refine_features(\n"
373" __read_only image2d_t grayscale,\n"
374" __global const float *harris_buf,\n"
375" __global float2 *refined_features,\n"
376" int subpixel_refine\n"
378" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
379" // The location in the grayscale buffer rather than the compacted grid\n"
380" int2 loc_i = (int2)(loc.x * 32, loc.y * 32);\n"
383" float max_val = 0;\n"
384" float2 loc_max = (float2)(-1, -1);\n"
386" int end_x = min(loc_i.x + 32, (int)get_image_dim(grayscale).x - 1);\n"
387" int end_y = min(loc_i.y + 32, (int)get_image_dim(grayscale).y - 1);\n"
389" for (int i = loc_i.x; i < end_x; ++i) {\n"
390" for (int j = loc_i.y; j < end_y; ++j) {\n"
391" new_val = harris_buf[i + j * get_image_dim(grayscale).x];\n"
393" if (new_val > max_val) {\n"
394" max_val = new_val;\n"
395" loc_max = (float2)(i, j);\n"
400" if (max_val == 0) {\n"
401" // There are no features in this part of the frame\n"
402" write_to_1d_arrf2(refined_features, loc, loc_max);\n"
406" if (subpixel_refine) {\n"
407" float mask[REFINE_WIN_H * REFINE_WIN_W];\n"
408" for (int i = 0; i < REFINE_WIN_H; i++) {\n"
409" float y = (float)(i - REFINE_WIN_HALF_H) / REFINE_WIN_HALF_H;\n"
410" float vy = exp(-y * y);\n"
412" for (int j = 0; j < REFINE_WIN_W; j++) {\n"
413" float x = (float)(j - REFINE_WIN_HALF_W) / REFINE_WIN_HALF_W;\n"
414" mask[i * REFINE_WIN_W + j] = (float)(vy * exp(-x * x));\n"
418" loc_max = corner_sub_pix(grayscale, loc_max, mask);\n"
421" write_to_1d_arrf2(refined_features, loc, loc_max);\n"
424"// Extracts BRIEF descriptors from the grayscale src image for the given features\n"
425"// using the provided sampler.\n"
426"__kernel void brief_descriptors(\n"
427" __read_only image2d_t grayscale,\n"
428" __global const float2 *refined_features,\n"
429" // for 512 bit descriptors\n"
430" __global ulong8 *desc_buf,\n"
431" __global const PointPair *brief_pattern\n"
433" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
434" float2 feature = read_from_1d_arrf2(refined_features, loc);\n"
436" // There was no feature in this part of the frame\n"
437" if (feature.x == -1) {\n"
438" write_to_1d_arrul8(desc_buf, loc, (ulong8)(0));\n"
443" ulong *p = &desc;\n"
445" for (int i = 0; i < 8; ++i) {\n"
446" for (int j = 0; j < 64; ++j) {\n"
447" PointPair pair = brief_pattern[j * (i + 1)];\n"
448" float l1 = read_imagef(grayscale, sampler_linear, feature + pair.p1).x;\n"
449" float l2 = read_imagef(grayscale, sampler_linear, feature + pair.p2).x;\n"
452" p[i] |= 1UL << j;\n"
457" write_to_1d_arrul8(desc_buf, loc, desc);\n"
460"// Given buffers with descriptors for the current and previous frame, determines\n"
461"// which ones match, writing correspondences to matches_buf.\n"
463"// Feature and descriptor buffers are assumed to be compacted (each element sourced\n"
464"// from a 32x32 block in the frame being processed).\n"
465"__kernel void match_descriptors(\n"
466" __global const float2 *prev_refined_features,\n"
467" __global const float2 *refined_features,\n"
468" __global const ulong8 *desc_buf,\n"
469" __global const ulong8 *prev_desc_buf,\n"
470" __global MotionVector *matches_buf\n"
472" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
473" ulong8 desc = read_from_1d_arrul8(desc_buf, loc);\n"
474" const int search_radius = 3;\n"
476" MotionVector invalid_vector = (MotionVector) {\n"
478" (float2)(-1, -1),\n"
484" if (desc.s0 == 0 && desc.s1 == 0) {\n"
485" // There was no feature in this part of the frame\n"
486" write_to_1d_arrvec(\n"
494" int2 start = max(loc - search_radius, 0);\n"
495" int2 end = min(loc + search_radius, (int2)(get_global_size(0) - 1, get_global_size(1) - 1));\n"
497" for (int i = start.x; i < end.x; ++i) {\n"
498" for (int j = start.y; j < end.y; ++j) {\n"
499" int2 prev_point = (int2)(i, j);\n"
500" int total_dist = 0;\n"
502" ulong8 prev_desc = read_from_1d_arrul8(prev_desc_buf, prev_point);\n"
504" if (prev_desc.s0 == 0 && prev_desc.s1 == 0) {\n"
508" ulong *prev_desc_p = &prev_desc;\n"
509" ulong *desc_p = &desc;\n"
511" for (int i = 0; i < 8; i++) {\n"
512" total_dist += popcount(desc_p[i] ^ prev_desc_p[i]);\n"
515" if (total_dist < DISTANCE_THRESHOLD) {\n"
516" write_to_1d_arrvec(\n"
521" read_from_1d_arrf2(prev_refined_features, prev_point),\n"
522" read_from_1d_arrf2(refined_features, loc)\n"
533" // There is no found match for this point\n"
534" write_to_1d_arrvec(\n"
541"// Returns the position of the given point after the transform is applied\n"
542"static float2 transformed_point(float2 p, __global const float *transform) {\n"
545" ret.x = p.x * transform[0] + p.y * transform[1] + transform[2];\n"
546" ret.y = p.x * transform[3] + p.y * transform[4] + transform[5];\n"
552"// Performs the given transform on the src image\n"
553"__kernel void transform(\n"
554" __read_only image2d_t src,\n"
555" __write_only image2d_t dst,\n"
556" __global const float *transform\n"
558" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
559" float2 norm = convert_float2(get_image_dim(src));\n"
566" sampler_linear_mirror,\n"
567" transformed_point((float2)(loc.x, loc.y), transform) / norm\n"
572"// Returns the new location of the given point using the given crop bounding box\n"
573"// and the width and height of the original frame.\n"
574"static float2 cropped_point(\n"
577" float2 bottom_right,\n"
582" float crop_width = bottom_right.x - top_left.x;\n"
583" float crop_height = bottom_right.y - top_left.y;\n"
585" float width_norm = p.x / (float)orig_dim.x;\n"
586" float height_norm = p.y / (float)orig_dim.y;\n"
588" ret.x = (width_norm * crop_width) + top_left.x;\n"
589" ret.y = (height_norm * crop_height) + ((float)orig_dim.y - bottom_right.y);\n"
594"// Upscales the given cropped region to the size of the original frame\n"
595"__kernel void crop_upscale(\n"
596" __read_only image2d_t src,\n"
597" __write_only image2d_t dst,\n"
599" float2 bottom_right\n"
601" int2 loc = (int2)(get_global_id(0), get_global_id(1));\n"
609" cropped_point((float2)(loc.x, loc.y), top_left, bottom_right, get_image_dim(dst))\n"
614"// Draws boxes to represent the given point matches and uses the given transform\n"
615"// and crop info to make sure their positions are accurate on the transformed frame.\n"
617"// model_matches is an array of three points that were used by the RANSAC process\n"
618"// to generate the given transform\n"
619"__kernel void draw_debug_info(\n"
620" __write_only image2d_t dst,\n"
621" __global const MotionVector *matches,\n"
622" __global const MotionVector *model_matches,\n"
623" int num_model_matches,\n"
624" __global const float *transform\n"
626" int loc = get_global_id(0);\n"
627" MotionVector vec = matches[loc];\n"
628" // Black box: matched point that RANSAC considered an outlier\n"
629" float4 big_rect_color = (float4)(0.1f, 0.1f, 0.1f, 1.0f);\n"
631" if (vec.should_consider) {\n"
632" // Green box: matched point that RANSAC considered an inlier\n"
633" big_rect_color = (float4)(0.0f, 1.0f, 0.0f, 1.0f);\n"
636" for (int i = 0; i < num_model_matches; i++) {\n"
637" if (vec.p.p2.x == model_matches[i].p.p2.x && vec.p.p2.y == model_matches[i].p.p2.y) {\n"
638" // Orange box: point used to calculate model\n"
639" big_rect_color = (float4)(1.0f, 0.5f, 0.0f, 1.0f);\n"
643" float2 transformed_p1 = transformed_point(vec.p.p1, transform);\n"
644" float2 transformed_p2 = transformed_point(vec.p.p2, transform);\n"
646" draw_box(dst, (int2)(transformed_p2.x, transformed_p2.y), big_rect_color, 5);\n"
647" // Small light blue box: the point in the previous frame\n"
648" draw_box(dst, (int2)(transformed_p1.x, transformed_p1.y), (float4)(0.0f, 0.3f, 0.7f, 1.0f), 3);\n"
const char * ff_source_deshake_cl